@@ -1,70 +1,48 @@
 # Copyright (c) Tumiz.
 # Distributed under the terms of the GPL-3.0 License.
 from __future__ import annotations
-import collections
-import uuid
 import numpy
-from IPython.display import display, update_display, HTML
+from IPython.display import HTML
+from typing import Dict
 import pathlib
-import json
 import uuid
+import json
 
 pi = numpy.arccos(-1)
 
 
-def sorted_sizes(*sizes):
-    return tuple(list(dict(sorted(collections.Counter(sizes).items(), key=lambda v: v[1])).keys()))
-
-
-def merge_shapes(*shapes):
-    ret = ()
-    ndim = max([len(s) for s in shapes])
-    for i in range(ndim):
-        sizes = []
-        for s in shapes:
-            if len(s) == 1 and s[-1] == 1:
-                continue
-            if len(s) > i:
-                sizes.append(s[-1-i])
-        ret = sorted_sizes(*sizes) + ret
-    return ret
-
-
-def force_assgin(v1, v2):
-    try:
-        numpy.copyto(v1, v2)
-    except:
-        if v1.size == v2.size:
-            v1[:] = v2.reshape(v1.shape)
-        else:
-            v1[:] = v2[:, numpy.newaxis]
+def sign(v):
+    return numpy.sign(v, where=v != 0, out=numpy.ones_like(v))
 
 
 class Viewer:
     tmp = open(pathlib.Path(__file__).parent/"viewer.html").read()
 
-    def __init__(self) -> None:
-        self.cache = {}
-        self.id = str(uuid.uuid1())
-        display(HTML(""), display_id=self.id)
+    def __init__(self, name="") -> None:
+        self.cache: Dict[float, list] = {}
+        self.id = name if name else str(uuid.uuid1())
 
-    def render_args(self, obj_id, t, **args):
+    def render_args(self, t, **args):
         if t in self.cache:
-            self.cache[t][obj_id] = args
+            self.cache[t].append(args)
         else:
-            self.cache[t] = {obj_id: args}
+            self.cache[t] = [args]
 
-    def show(self, debug=False):
+    def show(self, **args):
         html = self.tmp.replace("PY#D_ID", self.id).replace(
             "PY#D_ARGS", json.dumps(self.cache))
-        if debug:
-            open(self.id+".html", "w").write(html)
-        update_display(HTML(html), display_id=self.id)
+        if "name" in args:
+            open(args["name"]+".html", "w").write(html)
+        return HTML(html)
 
-    def render(self, *objs: Point, t=0, **args):
+    def render(self, *objs: Point, t=0):
         for obj in objs:
-            self.render_args(obj_id=id(obj), mode=obj.TYPE, t=t, vertice=obj.vertice.ravel(
-            ).tolist(), color=obj.color.ravel().tolist(), **args)
+            self.render_args(t=t, mode=obj.TYPE, vertex=obj.vertex.ravel(
+            ).tolist(), color=obj.color.ravel().tolist())
+
+    def label(self, text, position: list = [0, 0, 0], color="grey", t=0):
+        self.render_args(t=t, mode="TEXT", text=text,
+                         position=position, color=color)
 
 
 class Data(numpy.ndarray):
@@ -137,6 +115,22 @@ class Vector2(Vector):
     def __new__(cls, data: list | numpy.ndarray = [], n=()):
         return super().__new__(cls, data, n)
 
+    @property
+    def x(self):
+        return self[..., 0].view(numpy.ndarray)
+
+    @x.setter
+    def x(self, v):
+        self[..., 0] = v
+
+    @property
+    def y(self):
+        return self[..., 1].view(numpy.ndarray)
+
+    @y.setter
+    def y(self, v):
+        self[..., 1] = v
+
 
 class Vector3(Vector):
     BASE_SHAPE = 3,
@@ -145,18 +139,28 @@ class Vector3(Vector):
         if numpy.any(data):
             return super().__new__(cls, data, n)
         else:
-            x_: numpy.ndarray = numpy.array(x)
-            y_: numpy.ndarray = numpy.array(y)
-            z_: numpy.ndarray = numpy.array(z)
-            n += merge_shapes(x_.shape, y_.shape, z_.shape)
+            n += max(numpy.shape(x), numpy.shape(y), numpy.shape(z))
             ret = super().__new__(cls, [0., 0., 0.], n)
-            ret[..., 0] = x
-            ret[..., 1] = y
-            ret[..., 2] = z
+            ret.x = x
+            ret.y = y
+            ret.z = z
             return ret
 
+    @classmethod
+    def Grid(cls, x=0, y=0, z=0, n=()) -> Vector3:
+        n += numpy.shape(x) + numpy.shape(y) + numpy.shape(z)
+        ret = super().__new__(cls, [0., 0., 0.], n)
+        i = numpy.arange(len(n))
+        ret.x = numpy.expand_dims(x, axis=i[i != 0].tolist())
+        ret.y = numpy.expand_dims(y, axis=i[i != 1].tolist())
+        ret.z = numpy.expand_dims(z, axis=i[i != 2].tolist())
+        return ret
+
     def __matmul__(self, value: Transform) -> Vector3:
-        return numpy.matmul(self.H, value)[..., 0:3].view(self.__class__)
+        if type(value) is Transform:
+            return numpy.matmul(self.H, value)[..., 0:3].view(Vector3)
+        else:
+            return self.__matmul__(value)
 
     @property
     def x(self):
@@ -164,7 +168,7 @@ class Vector3(Vector):
 
     @x.setter
     def x(self, v):
-        force_assgin(self[..., 0], v)
+        self[..., 0] = v
 
     @property
     def y(self):
@@ -172,7 +176,7 @@ class Vector3(Vector):
 
     @y.setter
     def y(self, v):
-        force_assgin(self[..., 1], v)
+        self[..., 1] = v
 
     @property
     def z(self):
@@ -180,7 +184,7 @@ class Vector3(Vector):
 
     @z.setter
     def z(self, v):
-        force_assgin(self[..., 2], v)
+        self[..., 2] = v
 
     def dot(self, v) -> Vector3:
         if type(v) is Vector3:
@@ -247,14 +251,14 @@ class Vector3(Vector):
         f1 = self[i]
         return (1-d)*f0+d*f1
 
-    def as_scaling(self):
+    def as_scaling(self) -> Transform:
         ret = Transform(n=self.n)
         ret[..., 0, 0] = self[..., 0]
         ret[..., 1, 1] = self[..., 1]
         ret[..., 2, 2] = self[..., 2]
         return ret
 
-    def as_translation(self):
+    def as_translation(self) -> Transform:
         ret = Transform(n=self.n)
         ret[..., 3, 0] = self[..., 0]
         ret[..., 3, 1] = self[..., 1]
@@ -263,36 +267,93 @@ class Vector3(Vector):
 
     def as_point(self, color=None) -> Point:
         entity = Point(*self.n)
-        entity.vertice = self
+        entity.vertex = self
         if color is not None:
             entity.color = color
         return entity
 
-    def as_vector(self, start=0):
-        entity = Arrow(*self.n)
-        entity.start = start
-        entity.end = self
-        return entity
-
     def as_line(self) -> Line:
         entity = Line(*self.n)
-        entity.vertice = self
+        entity.vertex = self
         return entity
 
     def as_linesegment(self) -> LineSegment:
         entity = LineSegment(*self.n)
-        entity.vertice = self
+        entity.vertex = self
         return entity
 
-    def as_triangle(self):
+    def as_triangle(self) -> Triangle:
         entity = Triangle(*self.n[:-1])
-        entity.vertice = self
+        entity.vertex = self
         return entity
 
-    def as_mesh(self):
+    def as_mesh(self) -> Mesh:
         entity = Mesh.from_indexed(self)
         return entity
 
+    def as_vector(self) -> LineSegment:
+        entity = LineSegment(*self.n, 2)
+        entity.start.vertex = 0
+        entity.end.vertex = numpy.expand_dims(self, axis=self.ndim - 1)
+        return entity
+
+
+class Vector4(Vector):
+    BASE_SHAPE = 4,
+
+    def __new__(cls, xyzw_list: list | numpy.ndarray = [0., 0., 0., 1.], n=()):
+        return super().__new__(cls, xyzw_list, n)
+
+    @property
+    def x(self):
+        return self[..., 0].view(numpy.ndarray)
+
+    @x.setter
+    def x(self, v):
+        self[..., 0] = v
+
+    @property
+    def y(self):
+        return self[..., 1].view(numpy.ndarray)
+
+    @y.setter
+    def y(self, v):
+        self[..., 1] = v
+
+    @property
+    def z(self):
+        return self[..., 2].view(numpy.ndarray)
+
+    @z.setter
+    def z(self, v):
+        self[..., 2] = v
+
+    @property
+    def w(self):
+        return self[..., 3].view(numpy.ndarray)
+
+    @w.setter
+    def w(self, v):
+        self[..., 3] = v
+
+    @property
+    def xyz(self) -> Vector3:
+        return self[..., 0:3].view(Vector3)
+
+    @xyz.setter
+    def xyz(self, v):
+        self[..., 0:3] = v
+
+    @property
+    def wxyz(self) -> Vector:
+        ret = Vector([0, 0, 0, 0], n=self.n)
+        ret[..., 0] = self.w
+        ret[..., 1:4] = self.xyz
+        return ret
+
+    def as_transform(self) -> Transform:
+        return Transform.from_quaternion(self)
+
 
 class Transform(Data):
     BASE_SHAPE = 4, 4
@@ -354,35 +415,26 @@ class Transform(Data):
         return self @ self.Rz(a, n)
 
     @classmethod
-    def from_vector_change(cls, p0: Vector3, p1: Vector3, p0_: Vector3, p1_: Vector3) -> Transform:
-        d: Vector3 = p1 - p0
-        d_: Vector3 = p1_ - p0_
-        s = d_.L / d.L
-        ret = Transform(*s.shape)
-        angle = d.angle_to_vector(d_).squeeze()
-        axis = d.cross(d_)
-        ret.scaling = Vector3(s, s, s).squeeze()
-        ret.translation = p0_ - p0
+    def from_vector_change(cls, a: Vector3, b: Vector3) -> Transform:
+        ret = Transform(max(a.shape, b.shape))
+        angle = a.angle_to_vector(b).squeeze()
+        axis = a.cross(b)
         ret.rotation = cls.from_angle_axis(angle, axis)
         return ret
 
-    @classmethod
-    def from_angle_arbitrary_axis(cls, angle, axis_direction: Vector3, axis_point: Vector3) -> Transform:
-        return Transform.from_translation(-axis_point)@Transform.from_angle_axis(angle, axis_direction)@Transform.from_translation(axis_point)
-
     @classmethod
     def from_angle_axis(cls, angle, axis: list | Vector3) -> Transform:
         angle = numpy.array(angle)
         axis = Vector3(axis)
-        n = merge_shapes(angle.shape, axis.n)
+        n = max(angle.shape, axis.n)
         q = numpy.empty(n+(4,))
         half_angle = angle / 2
         q[..., 0] = numpy.cos(half_angle)
         q[..., 1:] = numpy.sin(half_angle)[..., numpy.newaxis] * axis.U
         return cls.from_quaternion(q)
 
-    def to_angle_axis(self):
-        q = self.to_quaternion()
+    def as_angle_axis(self):
+        q = self.as_quaternion()
         ha = numpy.arccos(q[..., 0])
         sin_ha = numpy.sin(ha)[..., numpy.newaxis]
         axis = numpy.divide(q[..., 1:], sin_ha,
@@ -390,34 +442,34 @@ class Transform(Data):
         return ha*2, axis
 
     @classmethod
-    def from_quaternion(cls, quaternion: list | numpy.ndarray) -> Transform:
-        q = numpy.array(quaternion)
-        w = q[..., 0]
-        x = q[..., 1]
-        y = q[..., 2]
-        z = q[..., 3]
-        ret = cls(*(q.shape[:-1]))
-        ret[..., 0, 0] = 1 - 2 * y ** 2 - 2 * z ** 2
-        ret[..., 0, 1] = 2 * w * z + 2 * x * y
-        ret[..., 0, 2] = -2 * w * y + 2 * x * z
-        ret[..., 1, 0] = -2 * w * z + 2 * x * y
-        ret[..., 1, 1] = 1 - 2 * x ** 2 - 2 * z ** 2
-        ret[..., 1, 2] = 2 * w * x + 2 * y * z
-        ret[..., 2, 0] = 2 * w * y + 2 * x * z
-        ret[..., 2, 1] = -2 * w * x + 2 * y * z
-        ret[..., 2, 2] = 1 - 2 * x ** 2 - 2 * y ** 2
+    def from_quaternion(cls, xyzw_list: list | numpy.ndarray) -> Transform:
+        q = Vector4(xyzw_list)
+        ret = cls(n=q.shape[:-1])
+        ret[..., 0, 0] = 1 - 2 * q.y ** 2 - 2 * q.z ** 2
+        ret[..., 0, 1] = 2 * q.w * q.z + 2 * q.x * q.y
+        ret[..., 0, 2] = -2 * q.w * q.y + 2 * q.x * q.z
+        ret[..., 1, 0] = -2 * q.w * q.z + 2 * q.x * q.y
+        ret[..., 1, 1] = 1 - 2 * q.x ** 2 - 2 * q.z ** 2
+        ret[..., 1, 2] = 2 * q.w * q.x + 2 * q.y * q.z
+        ret[..., 2, 0] = 2 * q.w * q.y + 2 * q.x * q.z
+        ret[..., 2, 1] = -2 * q.w * q.x + 2 * q.y * q.z
+        ret[..., 2, 2] = 1 - 2 * q.x ** 2 - 2 * q.y ** 2
         return ret
 
-    def to_quaternion(self) -> ndarray:
-        q = numpy.empty(self.n+(4,))
-        q[..., 0] = w = (1+self[..., 0, 0]+self[..., 1, 1] +
-                         self[..., 2, 2])**0.5/2
-        q[..., 1] = numpy.divide(
-            self[..., 1, 2]-self[..., 2, 1], 4*w, where=w != 0)
-        q[..., 2] = numpy.divide(
-            self[..., 2, 0]-self[..., 0, 2], 4*w, where=w != 0)
-        q[..., 3] = numpy.divide(
-            self[..., 0, 1]-self[..., 1, 0], 4*w, where=w != 0)
+    def as_quaternion(self) -> Vector4:
+        q = Vector4(n=self.n)
+        q.w = numpy.sqrt(1+self[..., 0, 0]+self[..., 1, 1] + self[..., 2, 2])/2
+        m0 = self[q.w == 0]
+        m1, w1 = self[q.w != 0], q.w[q.w != 0]
+        q.x[q.w != 0] = numpy.divide(m1[..., 1, 2]-m1[..., 2, 1], 4*w1)
+        q.y[q.w != 0] = numpy.divide(m1[..., 2, 0]-m1[..., 0, 2], 4*w1)
+        q.z[q.w != 0] = numpy.divide(m1[..., 0, 1]-m1[..., 1, 0], 4*w1)
+        q.x[q.w == 0] = sign(m0[..., 1, 2]-m0[..., 2, 1]) * numpy.sqrt(
+            1+m0[..., 0, 0]-m0[..., 1, 1] - m0[..., 2, 2])/2
+        q.y[q.w == 0] = sign(m0[..., 2, 0]-m0[..., 0, 2]) * numpy.sqrt(
+            1-m0[..., 0, 0]+m0[..., 1, 1] - m0[..., 2, 2])/2
+        q.z[q.w == 0] = sign(m0[..., 0, 1]-m0[..., 1, 0]) * numpy.sqrt(
+            1-m0[..., 0, 0]-m0[..., 1, 1] + m0[..., 2, 2])/2
         return q
 
     @classmethod
@@ -430,7 +482,7 @@ class Transform(Data):
         else:
             return m[lo[2]](v[..., 2]) @ m[lo[1]](v[..., 1]) @ m[lo[0]](v[..., 0])
 
-    def to_euler(self, sequence: str):
+    def as_euler(self, sequence: str):
         extrinsic = sequence.islower()
         lo = sequence.lower()
         ret = numpy.zeros((*self.n, 3))
@@ -458,8 +510,8 @@ class Transform(Data):
     def from_rpy(cls, angles_list: list | numpy.ndarray) -> Transform:
         return cls.from_euler('XYZ', angles_list)
 
-    def to_rpy(self):
-        return self.to_euler('XYZ')
+    def as_rpy(self):
+        return self.as_euler('XYZ')
 
     @property
     def translation_vector(self) -> Vector3:
@@ -509,10 +561,6 @@ class Transform(Data):
     def I(self) -> Transform:
         return numpy.linalg.inv(self)
 
-    @property
-    def forward(self) -> Vector3:
-        return Vector3(x=1).mt(self)
-
     @classmethod
     def from_perspective(cls, fovy, aspect, near, far) -> Transform:
         f = 1 / numpy.tan(fovy/2)
@@ -541,7 +589,7 @@ class Transform(Data):
         t1: Vector3 = self.translation_vector[i]
         s0: Vector3 = self.scaling_vector[i-1]
         s1: Vector3 = self.scaling_vector[i]
-        angle, axis = (r0.I@r1).to_angle_axis()
+        angle, axis = (r0.I@r1).as_angle_axis()
         rotation = Transform.from_angle_axis(d*angle, axis)
         translation = (d[..., numpy.newaxis] * (t1 - t0)).as_translation()
         scaling = (d[..., numpy.newaxis] * s1 / s0).as_scaling()
@@ -555,22 +603,14 @@ class Color(Vector):
         if numpy.any(data):
             return super().__new__(cls, data, n)
         else:
-            r_: numpy.ndarray = numpy.array(r)
-            g_: numpy.ndarray = numpy.array(g)
-            b_: numpy.ndarray = numpy.array(b)
-            a_: numpy.ndarray = numpy.array(a)
-            n += merge_shapes(r_.shape, g_.shape, b_.shape, a_.shape)
+            n += max(numpy.shape(r), numpy.shape(g),
+                     numpy.shape(b), numpy.shape(a))
             ret = super().__new__(cls, [0, 0, 0, 1], n)
             ret[..., 0] = r
             ret[..., 1] = g
             ret[..., 2] = b
             ret[..., 3] = a
-        return ret.view(cls)
-
-    def __len__(self):
-        if self.ndim == 1:
-            return 1
-        return super().__len__()
+        return ret
 
     @property
     def r(self):
@@ -578,7 +618,7 @@ class Color(Vector):
 
     @r.setter
     def r(self, v):
-        force_assgin(self[..., 0], v)
+        self[..., 0] = v
 
     @property
     def g(self):
@@ -586,7 +626,7 @@ class Color(Vector):
 
     @g.setter
     def g(self, v):
-        force_assgin(self[..., 1], v)
+        self[..., 1] = v
 
     @property
     def b(self):
@@ -594,7 +634,7 @@ class Color(Vector):
 
     @b.setter
     def b(self, v):
-        force_assgin(self[..., 2], v)
+        self[..., 2] = v
 
     @property
     def a(self):
@@ -616,11 +656,11 @@ class Point(Data):
         return ret
 
     @property
-    def vertice(self) -> Vector3:
+    def vertex(self) -> Vector3:
         return self[..., 0:3].view(Vector3)
 
-    @vertice.setter
-    def vertice(self, v):
+    @vertex.setter
+    def vertex(self, v):
         self[..., 0:3] = v
 
     @property
@@ -634,7 +674,8 @@ class Point(Data):
     def render(self, **args):
         v = Viewer()
         v.render(self, **args)
-        v.show()
+        return v.show(**args)
+
 
 class Triangle(Point):
     BASE_SHAPE = 3, 7
@@ -654,19 +695,11 @@ class LineSegment(Point):
 
     @property
     def start(self):
-        return self.vertice[..., ::2, :].squeeze().view(Vector3)
-
-    @start.setter
-    def start(self, v):
-        self.vertice[..., ::2, :].squeeze()[:] = v
+        return self[..., ::2, :].view(Point)
 
     @property
     def end(self):
-        return self.vertice[..., 1::2, :].squeeze().view(Vector3)
-
-    @end.setter
-    def end(self, v):
-        self.vertice[..., 1::2, :].squeeze()[:] = v
+        return self[..., 1::2, :].view(Point)
 
 
 class Line(Point):
@@ -687,7 +720,7 @@ class Mesh(Point):
     @classmethod
     def from_indexed(cls, v):
         ret = cls(*v.n)
-        ret.vertice = v
+        ret.vertex = v
         return ret
 
 
@@ -704,81 +737,6 @@ class Camera:
         return self.transform.I @ self.projection
 
 
-class Tetrahedron(Mesh):
-    def __init__(self, *n):
-        super().__init__(*n, 4)
-        self.index = [0, 1, 2, 0, 2, 3, 0, 1, 3, 1, 2, 3]
-
-
-class Cube(Mesh):
-    vertice_base = Vector3([
-        [0.5, 0.5, 0.5],
-        [-0.5, 0.5, 0.5],
-        [0.5, -0.5, 0.5],
-        [-0.5, -0.5, 0.5],
-        [0.5, 0.5, -0.5],
-        [-0.5, 0.5, -0.5],
-        [0.5, -0.5, -0.5],
-        [-0.5, -0.5, -0.5]])
-
-    def __init__(self, *n):
-        super().__init__(*n, 8)
-        self.geometry.vertice = self.vertice_base
-        self.index = [3,
-                      2,
-                      0,
-                      3,
-                      0,
-                      1,
-                      5,
-                      0,
-                      1,
-                      5,
-                      4,
-                      0,
-                      5,
-                      3,
-                      1,
-                      5,
-                      3,
-                      7,
-                      6,
-                      2,
-                      0,
-                      6,
-                      4,
-                      0,
-                      6,
-                      3,
-                      2,
-                      6,
-                      3,
-                      7,
-                      6,
-                      5,
-                      4,
-                      6,
-                      5,
-                      7]
-
-
-class Arrow(LineSegment):
-    head_base = Vector3(
-        [
-            [1, 0, 0],
-            [0.9, 0, 0.05],
-            [0.9, -0.05 * numpy.cos(pi / 6), -0.05 * numpy.sin(pi / 6)],
-            [0.9, 0.05 * numpy.cos(pi / 6), -0.05 * numpy.sin(pi / 6)],
-        ]
-    )
-
-    def __new__(cls, *n):
-        ret = super().__new__(cls, *n, 2).view(cls)
-        ret.head = Tetrahedron(*n)
-        ret.head.color = ret.color[..., 0, numpy.newaxis, :]
-        return ret
-
-
 class Utils:
     car_points: Vector3 = Vector3.load(pathlib.Path(__file__).parent/"car.npy")
 
@@ -802,19 +760,25 @@ class Utils:
                 [-1, -1, 0]
             ])
             index = [0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 0, 6, 7, 7, 8, 8, 9, 9,
-                    10, 10, 11, 11, 12, 12, 6, 0, 7, 1, 8, 2, 9, 3, 10, 4, 11, 5, 12]
+                     10, 10, 11, 11, 12, 12, 6, 0, 7, 1, 8, 2, 9, 3, 10, 4, 11, 5, 12]
             return ret[index].as_linesegment()
         else:
             return cls.car_points.as_point()
 
     @classmethod
     def Grid(cls, size=5) -> LineSegment:
-        point: Vector3 = Vector3(x=[-size, size], y=-size)
-        point @= Transform.from_translation(y=range(2*size+1))
-        point = point.reshape(point.size//3, 3)
-        point @= Transform.from_rpy([[0, 0, 0], [0, 0, pi/2]])
-        grid = point.as_linesegment()
-        grid.color = Color(a=1)
-        grid[0].color[size*2:size*2+2] = Color(r=[0, 1])
-        grid[1].color[size*2:size*2+2] = Color(g=[0, 1])
-        return grid
+        v = Vector3(x=[-size, size]) @ Transform.from_translation(
+            y=range(-size, size+1)) @ Transform.from_rpy([[[0, 0, 0]], [[0, 0, pi/2]]])
+        l = v.as_linesegment()
+        l.color = Color()
+        l.color[0, size] = Color(r=[0, 1])
+        l.color[1, size] = Color(g=[0, 1])
+        return l
+
+    @classmethod
+    def Axis(cls, size=5) -> LineSegment:
+        a = Vector3([[size, 0, 0], [0, size, 0], [0, 0, size]]).as_vector()
+        a[0].color = Color(r=1)
+        a[1].color = Color(g=1)
+        a[2].color = Color(b=1)
+        return a
